import { join, resolve, basename } from 'path';
import { createPage } from '@vuepress/core';
import _docgen, { extractConfig } from 'vue-docgen-cli';
import WebpackConfig from 'webpack-chain';
import glob from 'globby';
import defu from 'defu';
import chokidar from 'chokidar';
import 'fs/promises';
import 'fs';
import matter from 'gray-matter';

const webpackHandleResolve = async ({
  app,
  config,
  isServer
}) => {
  config.resolve.alias.set("@source", app.dir.source()).set("@temp", app.dir.temp()).set("@internal", app.dir.temp("internal"));
  config.merge({
    resolve: {
      extensionAlias: {
        ".js": [".js", ".ts"],
        ".mjs": [".mjs", ".mts"]
      },
      extensions: [".js", ".jsx", ".ts", ".tsx", ".vue", ".json"]
    }
  });
  config.resolve.extensions.merge([
    ".js",
    ".jsx",
    ".ts",
    ".tsx",
    ".vue",
    ".json"
  ]);
  const aliasResult = await app.pluginApi.hooks.alias.process(app, isServer);
  aliasResult.forEach(
    (aliasObject) => Object.entries(aliasObject).forEach(([key, value]) => {
      config.resolve.alias.set(key, value);
    })
  );
};

const sleep = (ms = 0) => new Promise((res) => setTimeout(res, ms));

const extractAndCutFrontmatter = (doc, grayMatterOptions, content = "") => {
  const {
    data: topContentFrontmatter = {},
    content: cuttedTopContent
    // Without .trim() matter ignore some files
  } = matter(content.trim(), grayMatterOptions);
  let frontmatter = topContentFrontmatter;
  doc.docsBlocks?.forEach((blockContent, index) => {
    const {
      data,
      content: cuttedBlockContent
    } = matter(blockContent.trim(), grayMatterOptions);
    frontmatter = defu(data, frontmatter);
    doc.docsBlocks[index] = cuttedBlockContent;
  });
  return {
    content: matter.stringify(cuttedTopContent, frontmatter, grayMatterOptions),
    frontmatter
  };
};

const templateComponent = (grayMatterOptions, originalComponentTemplate) => (renderedUsage, doc, config, fileName, requiresMd, subTemplateOptions) => {
  if (subTemplateOptions.isSubComponent) {
    return originalComponentTemplate(
      renderedUsage,
      doc,
      config,
      fileName,
      requiresMd,
      subTemplateOptions
    );
  }
  const originalResultWithoutDocsBlocks = originalComponentTemplate(
    renderedUsage,
    {
      ...doc,
      docsBlocks: []
    },
    config,
    fileName,
    requiresMd,
    subTemplateOptions
  );
  const {
    content
  } = extractAndCutFrontmatter(
    doc,
    grayMatterOptions,
    originalResultWithoutDocsBlocks
  );
  return content;
};

const tmpFolderName = "docgen";

const docgen = _docgen.default;
const VueDocgenPlugin = ({
  docgenCliConfig = {},
  docgenCliConfigPath,
  pages = [{ components: ["**/components/**/*.vue", "!**/node_modules/**", "!**/.vuepress/**"] }]
}) => {
  if (!Array.isArray(pages))
    pages = [pages];
  const normalizedPages = pages.map((stringOrObject) => {
    if (typeof stringOrObject === "string")
      return {
        components: stringOrObject
      };
    return stringOrObject;
  });
  return {
    name: "vuepress-plugin-docgen",
    onInitialized: async (app) => {
      const { grayMatterOptions } = app.options.markdown.frontmatter || {};
      const tmpFolder = join(app.options.temp, tmpFolderName);
      const safeDocgenCliConfig = defu(docgenCliConfig, extractConfig(process.cwd(), app.env.isDev, docgenCliConfigPath, []));
      const webpackConfig = new WebpackConfig();
      await webpackHandleResolve({ app, config: webpackConfig, isServer: true });
      const baseDocgenCliConfig = {
        ...safeDocgenCliConfig,
        apiOptions: {
          jsx: true,
          ...safeDocgenCliConfig.apiOptions,
          ...webpackConfig.toConfig().resolve
        },
        templates: {
          ...safeDocgenCliConfig.templates,
          component: templateComponent(grayMatterOptions, safeDocgenCliConfig.templates.component),
          ...docgenCliConfig?.templates
        }
      };
      await Promise.all(normalizedPages.map(async ({
        root: componentsRoot,
        components,
        outDir: rawOutDir = ""
      }) => {
        const outDir = resolve(tmpFolder, rawOutDir);
        const config = {
          ...baseDocgenCliConfig,
          ...componentsRoot && { componentsRoot },
          components,
          outDir
        };
        await docgen(config);
      }));
      await sleep(100);
      const docFiles = await glob("**/*.md", {
        cwd: tmpFolder
      });
      await Promise.all(docFiles.map(async (relativeDocPath) => {
        let normalizedRelativeDocPath = relativeDocPath;
        const docBasename = basename(relativeDocPath);
        normalizedRelativeDocPath = normalizedRelativeDocPath.slice(0, -3);
        if (docBasename.toLowerCase() === "readme.md")
          normalizedRelativeDocPath = normalizedRelativeDocPath.replace(/readme$/i, "");
        if (docBasename.toLowerCase() === "index.md")
          normalizedRelativeDocPath = normalizedRelativeDocPath.replace(/index$/i, "");
        if (normalizedRelativeDocPath.endsWith("/"))
          normalizedRelativeDocPath = normalizedRelativeDocPath.slice(0, -1);
        const page = await createPage(app, {
          path: "/" + normalizedRelativeDocPath,
          filePath: join(tmpFolder, relativeDocPath),
          isDocgenPage: true
        });
        if (page.permalink)
          page.path = page.permalink;
        app.pages.push(page);
      }));
    },
    onWatched(app, watchers, restart) {
      const tmpFolder = join(app.options.temp, tmpFolderName);
      const watcher = chokidar.watch("**/*.md", {
        cwd: tmpFolder,
        ignoreInitial: true
      });
      watcher.on("change", () => restart());
      watchers.push(watcher);
    }
  };
};

export { VueDocgenPlugin, extractAndCutFrontmatter };
